#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114, C0115, C0116, E0401, R0201, R0903, W0613, W0621, W1203

from __future__ import annotations

import datetime
import json
import logging
import pathlib
import re
import subprocess
from typing import Dict, List, Optional

import chevron
import pendulum
import requests
import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import registry, relationship

from workspaces.capability.enums import (
    CapabilityRequestState,
    CapabilityVersionState,
    QueueState,
)
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.services.interfaces import ExecutionManagerIF
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import AbstractFile, JSONSerializable

mapper_registry = registry()

logger = logging.getLogger(__name__)

CAPABILITY_NAME_FK = "capabilities.capability_name"
CAPABILITY_SETTINGS_KEY = "edu.nrao.workspaces.CapabilitySettings"
WORKFLOW_SETTINGS_KEY = "edu.nrao.workspaces.WorkflowSettings"
NOTIF_SETTINGS_KEY = "edu.nrao.workspaces.NotificationSettings"


@mapper_registry.mapped
class Action:
    __tablename__ = "capability_state_actions"
    action_id = sa.Column(
        "action_id",
        sa.Integer,
        primary_key=True,
        autoincrement=True,
        comment="The unique identifier for this state action",
    )
    transition_id = sa.Column(
        "transition_id",
        sa.Integer,
        sa.ForeignKey("capability_state_transitions.transition_id"),
        nullable=False,
        comment="The transition this action is executed during",
    )
    action_type = sa.Column(
        "action_type",
        sa.String,
        nullable=False,
        comment="The type of action this is. Most likely the name of an Action subclass such as SendNotification or ExecuteWorkflow",
    )
    arguments = sa.Column("arguments", sa.String, comment="Additional arguments for the action")
    transition = relationship("Transition", back_populates="actions")
    __mapper_args__ = {"polymorphic_on": action_type}


@mapper_registry.mapped
class Noop(Action):
    __mapper_args__ = {"polymorphic_identity": "Noop"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        # empty because this is the no-operation action
        pass


@mapper_registry.mapped
class ExecuteWorkflow(Action):
    __mapper_args__ = {"polymorphic_identity": "ExecuteWorkflow"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        logger.info("Executing workflow %s for execution %s", self.arguments, execution.id)
        workflow_name = self.arguments
        workflow_args = None
        parameters = execution.version.parameters
        files = execution.version.files
        if parameters is not None:
            # convert to json string because json cares about single vs. double quotes.
            # DO NOT TAKE THIS OUT! Python will yell at you.
            parameters = json.dumps(parameters)
            workflow_args = json.loads(parameters)
        logger.info(
            "submitting workflow request with args %s (and type %s)",
            workflow_args,
            type(workflow_args),
        )
        manager.submit_workflow_request(execution.id, workflow_name, workflow_args, files)


@mapper_registry.mapped
class QueueWorkflow(Action):
    __mapper_args__ = {"polymorphic_identity": "QueueWorkflow"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        # message the execution manager
        logger.info(f"Enqueuing workflow {self.arguments}")
        manager.queue_execution(execution)


@mapper_registry.mapped
class SendNotification(Action):
    __mapper_args__ = {"polymorphic_identity": "SendNotification"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        #
        # Perform any necessary preparatory work (including email overrides and a check
        # of CAPO, in order to send appropriate emails via the notification service.
        # ...

        args = json.loads(self.arguments)

        # Check if sending to DAs or user
        send_to_aod = "send_to_aod" in args and args["send_to_aod"]
        send_to_das = "send_to_qa" in args and args["send_to_qa"]
        obtain_contacts = "needs_contacts" in args and args["needs_contacts"]
        if execution.version.parameters is not None:
            send_to_user = "user_email" in execution.version.parameters
        else:
            send_to_user = False

        # Set the destination email
        if send_to_das:
            # If this flag is set, overwrite any provided user email address:
            # Check for an assigned DA, and send to them if you can, otherwise use the lsit
            dest_email = self.obtain_da_email(execution)
        elif send_to_aod:
            # Get the assigned AoD's email, or the address of available AoDs as a backup
            dest_email = self.find_aod_email(execution)
        elif obtain_contacts:
            # If this flag is set, we expect no user_email information.
            # Obtain the list of recipients (or the DA list if not production)
            dest_email = self.obtain_contacts(execution.version)
        elif send_to_user:
            # If neither of the above apply, send it to the provided email address.
            dest_email = execution.version.parameters["user_email"]
        else:
            logger.debug("Sending no email notification because there is no user_email in the version parameters")
            return

        # It's possible lookups might fail:
        if dest_email is None:
            logger.error("Could not obtain an address, no email being sent")
            return

        # Announce where things are being sent:
        logger.info(f">> Sending to user at {dest_email}")

        # Make sure we've been given a template, but it's almost guaranteed:
        if "template" not in args:
            logger.debug("Sending no email notification because there was not template specified")
        else:
            # Build up the JSON to provide to the notification service, and add the decided upon email.
            notification_parameters = {"version": execution.version.__json__(), "destination_email": dest_email}
            logger.debug("Preparing to send a notification to %s", dest_email)
            manager.notifier.send_email(args["template"], notification_parameters)

    @staticmethod
    def obtain_da_email(execution):
        # fallback to the list
        dest_email = CapoConfig().settings(NOTIF_SETTINGS_KEY).analystEmail
        # but overwrite it if we have an assigned DA's email
        if execution.capability_request.assigned_da is not None:
            if execution.capability_request.da.email is not None:
                dest_email = execution.capability_request.da.email
        return dest_email

    @staticmethod
    def find_aod_email(execution):
        # Grab the email for the assigned AoD, or don't send anything.
        if execution.capability_request.assigned_aod is not None:
            dest_email = execution.capability_request.aod.email
        else:
            dest_email = None
        return dest_email

    def obtain_contacts(self, version: CapabilityVersion):
        # Are we sending emails to the PI?  Only for production, and not always then.
        email_pi = CapoConfig().getboolean("edu.nrao.workspaces.NotificationSettings.sendPIEmail")

        if email_pi:
            # Call the AAT and PST wrest tools in order to acquire the contact list
            email_list = self.call_wresters(version)

            if isinstance(email_list, list):
                send_to = ", ".join(email_list)
            else:
                send_to = email_list

            logger.info(f"Email to be sent to project contacts: {send_to}")
        else:
            logger.info("User email overridden to DA list via CAPO setting.")
            # Otherwise, we send the email to the analysts list:
            send_to = CapoConfig().settings(NOTIF_SETTINGS_KEY).analystEmail

        return send_to

    @staticmethod
    def call_wresters(version):
        logger.info(f"Wresting contact info for PI & CoIs of request #{version.capability_request_id}")
        if version.parameters is not None:
            if "product_locator" in version.parameters:
                process_call = ["aat_wrest", "-sc", version.parameters["product_locator"]]
            elif "sdmId" in version.parameters:
                process_call = ["aat_wrest", "-si", version.parameters["sdmId"]]
            else:
                logger.debug(f"Neither SDM ID nor Product Locator available!  Exiting without calling wresters")
                return None
        else:
            logger.debug("Empty parameters, could not obtain a starting point for initiating wrests")
            return None

        with subprocess.Popen(
            process_call,
            stdout=subprocess.PIPE,
        ) as project_data:
            data = project_data.communicate()[0]
            project = json.loads(data)
            project_code = project["projectCode"]

            with subprocess.Popen(
                ["contacts_wrest", project_code],
                stdout=subprocess.PIPE,
            ) as wrested_contacts:
                contacts = wrested_contacts.communicate()[0]
                if "ERROR" in contacts.decode():
                    logger.info(f"Not sending notification because contact wrester failed for project {project_code}")
                    return None

        # So far, so good, transform the output and continue on
        email_list = json.loads(contacts.decode())
        return email_list


@mapper_registry.mapped
class SendMessage(Action):
    """
    Action that simply sends an AMQP message with the type of the given arguments
    """

    __mapper_args__ = {"polymorphic_identity": "SendMessage"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        message_type = self.arguments
        message_architect = CapabilityMessageArchitect(execution=execution)
        message_to_send = message_architect.compose_message(message_type)
        logger.info("Sending %s message regarding execution #%s", message_to_send["type"], execution.id)
        manager.messenger.send_message(**message_to_send)


@mapper_registry.mapped
class QaPass(Action):
    """
    Action that performs system-level operations required by QA-passing a capability version
    """

    __mapper_args__ = {"polymorphic_identity": "QaPass"}

    def _fail_other_versions(self, request: CapabilityRequest, pass_version: CapabilityVersion):
        """
        Given a capability request and version to pass, set the state of the passed version to Complete,
        and the state of all other versions to Failed

        :param request: Capability request
        """
        versions = request.versions
        for version in versions:
            if version.__json__() != pass_version.__json__():
                # Mark version as failed
                print(f"Failing version: {version.__json__()}")
                if self.check_still_running(version):
                    self.abort_running_execution(version)
                else:
                    # if QAFail has not already been run, or this execution was aborted, QAFail it
                    if version.state != CapabilityVersionState.Failed.name:
                        self.run_qa_fail_workflow(version)
                version.state = CapabilityVersionState.Failed.name

    @staticmethod
    def check_still_running(version: CapabilityVersion) -> bool:
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        if version.current_execution.current_workflow_request_id:
            response = requests.get(
                f"{workflow_service_url}/workflows/{version.capability.name}/requests/"
                f"{version.current_execution.current_workflow_request_id}/htcondor_id"
            )
            if json.loads(response.content.decode())["htcondor_job_id"] == "None":
                return False
            return True
        else:
            return False

    @staticmethod
    def run_qa_fail_workflow(version: CapabilityVersion):
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        logger.info(f"QA Failing version: {version.__json__()}")
        if version.current_execution.current_workflow_request_id:
            requests.post(
                f"{workflow_service_url}/workflows/{version.capability.name}/requests/"
                f"{version.current_execution.current_workflow_request_id}/qa_fail"
            )

    @staticmethod
    def abort_running_execution(version: CapabilityVersion):
        """
        Given a capability version, fail all running executions

        :param version: Capability version
        """
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        # There is only a single active execution for any version
        workflow_request_id = version.current_execution.current_workflow_request_id
        logger.info(f"Attempting to abort running execution #{version.current_execution.id}.")
        requests.post(f"{workflow_service_url}/workflows/requests/{workflow_request_id}/abort")

    @staticmethod
    def _run_ingestion_workflow(execution: CapabilityExecution):
        """
        Start ingestion workflow for given execution

        :param execution: Given execution
        """
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        wf_req_id = execution.current_workflow_request_id

        response = requests.post(
            f"{workflow_service_url}/workflows/{execution.capability_name}/requests/{wf_req_id}/ingest"
        )
        # In case something goes wrong:
        response.raise_for_status()

    def __call__(self, execution: CapabilityExecution, *args):
        logger.info("Execution #%s has passed QA", execution.id)
        request = execution.capability_request
        pass_version = execution.version
        self._fail_other_versions(request, pass_version)
        self._run_ingestion_workflow(execution)


@mapper_registry.mapped
class QaFail(Action):
    """
    Action that performs system-level operations required by QA-failing a capability version
    """

    __mapper_args__ = {"polymorphic_identity": "QaFail"}

    def __call__(self, execution: CapabilityExecution, *args):
        """
        Launch qa-fail workflow for execution's workflow request and fail associated version

        :param execution: Execution being qa-failed
        """
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        wf_req_id = execution.current_workflow_request_id

        response = requests.post(
            f"{workflow_service_url}/workflows/{execution.capability_name}/requests/{wf_req_id}/qa_fail"
        )
        # just in case something goes wrong
        response.raise_for_status()
        # Fail version associated with execution
        execution.version.state = CapabilityVersionState.Failed.name


@mapper_registry.mapped
class QaAbandon(Action):
    """
    Action that performs all system-level operations required by the QA-failing of an
    entire capability request (as opposed to QaFail, which handles an individual capability
    version).
    """

    __mapper_args__ = {"polymorphic_identity": "QaAbandon"}

    @staticmethod
    def _fail_all_versions(request: CapabilityRequest):
        """
        Variation on the bookkeeping used by QaPass, which will set ALL versions to
        failed, and initiate the relevant workflow.

        :param request:
        :return:
        """
        versions = request.versions
        for version in versions:
            # Mark version as failed
            logger.debug(f"Failing version: {version.__json__()}")
            if QaPass.check_still_running(version):
                QaPass.abort_running_execution(version)
            else:
                # if QAFail has not already been run, or this execution was aborted, QAFail it
                if version.state != CapabilityVersionState.Failed.name:
                    logger.debug(f"Executing QaFail workflow for: {version.__json__()}")
                    QaPass.run_qa_fail_workflow(version)
            version.state = CapabilityVersionState.Failed.name

    @staticmethod
    def _flag_execution_block(request: CapabilityRequest):
        """
        Use the workflow server's endpoint for communicating with the AAT/PPI for
        calibration_status updates on the SDM associated with a std_calibration
        workflow

        :param request:
        :return:
        """
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl

        # wf_request_id
        request_id = request.current_version.current_execution.current_workflow_request_id

        # Call REST endpoint that shoots off QA Fail message
        url = f"{workflow_service_url}/workflows/requests/{request_id}/send-do-not-calibrate"
        response = requests.post(url)

        # Expect smooth sailing, but tell us details if something goes wrong
        response.raise_for_status()

    def __call__(self, execution: CapabilityExecution, *args):
        """
        Similar to QaPass in that all versions are resolved and conveyed, but in
        this case none of the versions are considered acceptable.

        This is also the point at which the Do Not Calibrate status is set in the
        AAT database.

        :param execution:
        :param args:
        :return:
        """

        logger.info(f"Cleaning up all versions of Request: {execution.capability_request}")

        # Clean up all versions:
        self._fail_all_versions(execution.capability_request)
        # Flag the EB in the AAT/PPI
        self._flag_execution_block(execution.capability_request)

        # Lastly, Seal the request:
        execution.capability_request.update_state(CapabilityRequestState.Failed)
        execution.capability_request.update_sealed(True)


@mapper_registry.mapped
class AnnounceQa(Action):
    """
    Action that makes a REST call to the announce_qa endpoint in the workflow service, which announces that an execution
    is ready for QA
    """

    __mapper_args__ = {"polymorphic_identity": "AnnounceQa"}

    def __call__(self, execution: CapabilityExecution, *args):
        workflow_request_id = execution.current_workflow_request_id
        msg_type = "qa_ready"
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        url = f"{workflow_service_url}/workflows/requests/{workflow_request_id}/qa/{msg_type}"
        requests.post(url, json={"capability_version": execution.version_number})


@mapper_registry.mapped
class Cancel(Action):
    """
    Action that makes a REST call to the abort workflow endpoint in the workflow service, which kills the htcondor job for a given workflow
    """

    __mapper_args__ = {"polymorphic_identity": "Cancel"}

    def __call__(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        workflow_request_id = execution.current_workflow_request_id
        logger.info("Cancelling workflow %s for execution %s", workflow_request_id, execution.id)
        workflow_service_url = CapoConfig().settings(WORKFLOW_SETTINGS_KEY).serviceUrl
        url = f"{workflow_service_url}/workflows/requests/{workflow_request_id}/abort"
        requests.post(url, json={"capability_version": execution.version_number})


@mapper_registry.mapped
class State:
    transitions: List[Transition]

    __tablename__ = "capability_states"

    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey(CAPABILITY_NAME_FK),
        primary_key=True,
        comment="The capability this state belongs to",
    )
    state_name = sa.Column(
        "state",
        sa.String,
        nullable=False,
        primary_key=True,
        comment="A state in this state machine, such as 'Executing' or 'Awaiting QA'",
    )

    def link(self, pattern: str, state: State, action: Action):
        transition = Transition(pattern=pattern, from_state=self, to_state=state)
        transition.actions.append(action)
        self.transitions.append(transition)

    def find_matching_transition(self, message: Dict) -> Optional[Transition]:
        """
        Find matching transition for state based on a given message

        :param message: Message to match transition with
        :return: Matching transition if one is found; else None
        """
        for transition in self.transitions:
            if transition.matches(message):
                return transition
        return None

    def get_next_state(self, message: Dict) -> State:
        """
        Get next state for current transition, given a message

        :param message: Message to match transition with
        :return: Next state if matching transition is found; else current state
        """
        transition = self.find_matching_transition(message)
        if transition:
            return transition.to_state
        return self

    def perform_action(self, message: Dict, execution: CapabilityExecution, manager: ExecutionManagerIF):
        """
        Perform action for the current transition, if one exists

        :param message: Message to match transition with
        :param execution: Execution that is performing the action
        :param manager: Execution manager to send to the action
        """
        transition = self.find_matching_transition(message)
        if transition:
            transition.action(execution, manager)

    def final(self) -> bool:
        return not self.transitions

    def __str__(self):
        return self.state_name


@mapper_registry.mapped
class QaStaff(JSONSerializable):
    """
    A staff member capable of performing QA actions
    """

    user_name: str
    group: str
    available: bool
    email: str

    __tablename__ = "qa_staff"

    user_name = sa.Column(
        "user_name",
        sa.String,
        primary_key=True,
        comment="The user name of the staff member performing QA.",
    )
    group = sa.Column(
        "group",
        sa.String,
        comment="Is this staff member a DA or an AOD?",
    )
    available = sa.Column(
        "available",
        sa.Boolean,
        default=True,
        comment="Is this staff member available to be assigned QA jobs?",
    )
    email = sa.Column(
        "email",
        sa.String,
        default=None,
        comment="Contact Email for the staff member",
    )

    @property
    def is_available(self):
        return self.available

    def __json__(self, request=None) -> dict:
        return {
            "user_name": self.user_name,
            "group": self.group,
            "available": self.available,
            "email": self.email,
        }


@mapper_registry.mapped
class Capability(JSONSerializable):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """

    id: int
    name: str
    max_jobs: int
    templates: List["CapabilityTemplate"]

    __tablename__ = "capabilities"
    name = sa.Column("capability_name", sa.String, primary_key=True)
    start_state = sa.Column(
        "start_state",
        sa.String,
        # Magic string required here, I think  -nhertz
        default="Start",
        server_default="Start",
    )
    max_jobs = sa.Column("max_jobs", sa.Integer)
    enabled = sa.Column("enabled", sa.Boolean, default=True, server_default="true")
    paused = sa.Column("paused", sa.Boolean, default=False, server_default="false")
    has_image_products = sa.Column(
        "has_image_products", sa.Boolean, default=False, server_default="false", nullable=False
    )
    requires_qa = sa.Column("requires_qa", sa.Boolean, default=False)
    single_version_only = sa.Column("single_version_only", sa.Boolean, default=False)
    templates = relationship("CapabilityTemplate", backref="capabilities", lazy="dynamic")
    requests = relationship("CapabilityRequest", back_populates="capability")

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF):
        """
        Create a new request for this capability with specific options

        FIXME: Future products needs a string representation
        """
        request = CapabilityRequest(capability=self, parameters=str(parameters))
        self.requests.append(request)
        return request

    def render_templates(self, arguments: Dict) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow

        :param arguments: the capability arguments JSON
        :return: a list of rendered template files
        """
        rendered_files = []

        for template in self.templates:
            rendered_files.append(template.render(arguments))

        return rendered_files

    def __str__(self) -> str:
        return f"Capability: {self.name}, max concurrent jobs of {self.max_jobs}"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "name": self.name,
            "max_jobs": self.max_jobs,
            "enabled": self.enabled,
            "has_image_products": self.has_image_products,
            "requires_qa": self.requires_qa,
            "single_version_only": self.single_version_only,
            "paused": self.paused,
        }

    @classmethod
    def from_json(cls, json: dict) -> Capability:
        """
        Method that creates an instance of a Capability from given JSON

        :param json: the description of this capabiity in a dictionary
        :return: Capability instance
        """
        copy = dict(json)
        del copy["type"]
        return Capability(**copy)


@mapper_registry.mapped
class CapabilityTemplate(JSONSerializable):
    """
    Capability Templates are Mustache-formatted files associated with a given Capability
    These are the templates that will be made available to the users for editing
    """

    filename: str
    content: bytes

    __tablename__ = "capability_templates"
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey("capabilities.capability_name"),
        primary_key=True,
    )

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Capability Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """
        contents = chevron.render(self.content.decode("ascii"), argument)
        logger.info(
            "rendering %s with argument %s to %s",
            self.content.decode("ascii"),
            argument,
            contents,
        )
        return AbstractFile(self.filename, contents.encode("ascii"))

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        # Defer to the actual template file's contents for JSON conversion
        return {
            "filename": self.filename,
            "content": self.content.decode("utf8"),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return CapabilityTemplate(json["filename"], json["content"])

    def __repr__(self):
        return f"<CapabilityTemplate filename={self.filename} for capability={self.capability_name}>"


@mapper_registry.mapped
class Transition:
    pattern: str
    actions: List[Action]
    from_state: State
    to_state: State

    __tablename__ = "capability_state_transitions"
    transition_id = sa.Column(
        "transition_id",
        sa.Integer,
        primary_key=True,
        autoincrement=True,
        comment="The unique identifier of this state machine transition",
    )

    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey(Capability.name),
        nullable=False,
        comment="The capability to whose state machine this transition belongs",
    )

    from_state_name = sa.Column(
        "from_state",
        sa.String,
        nullable=False,
        comment="The starting state for this transition",
    )

    pattern = sa.Column(
        "pattern",
        sa.String,
        nullable=False,
        comment="A JSONPATH pattern for what events cause us to take this transition",
    )

    to_state_name = sa.Column(
        "to_state",
        sa.String,
        nullable=False,
        comment="The destination state for this transition",
    )

    from_state = relationship(State, foreign_keys=[capability_name, from_state_name], backref="transitions")

    # SQLAlchemy wants to produce a warning here to inform you that if your from-state and your to-state belong to
    # different capabilities, then one of these relationships will overwrite that field for the other and your
    # application will get confused. Suppose we transition from "executing" of std-imaging to "await QA" of
    # std-calibration. There's only one slot here for a capability to be named.
    #
    # Of course, we won't let that happen in practice, so the "overlaps" message here basically says
    # "ignore that problem, the application isn't going to allow it to happen."
    to_state = relationship(State, foreign_keys=[capability_name, to_state_name], overlaps="from_state,transitions")

    actions = relationship(Action, back_populates="transition")

    __table_args__ = (
        sa.ForeignKeyConstraint(
            (capability_name, from_state_name),
            [State.capability_name, State.state_name],
        ),
        sa.ForeignKeyConstraint((capability_name, to_state_name), [State.capability_name, State.state_name]),
        {},
    )

    def matches(self, message: Dict) -> bool:
        """
        Check if a message matches a given pattern

        Patterns are stored (in the database) in the following format:
        {key] == {value}

        Only one pattern per transition is supported currently

        :param message: Message to be checked for matches
        :return: True if there's a match; else False
        """
        key, value, *rest = re.split(r"\s*==\s*", self.pattern)
        if len(rest) or not value:
            logger.info(
                f"Pattern {self.pattern} is malformed. It should take the form '{{key}} == {{value}}', "
                f"corresponding to the key and value of the incoming message."
            )
        logger.info(
            "Checking to see if key %s has value %s (it is %s)", key, value, message[key] if key in message else "None"
        )
        return key in message.keys() and value == message[key]

    def action(self, execution: CapabilityExecution, manager: ExecutionManagerIF):
        logger.info("Processing my actions")
        for action in self.actions:
            logger.info("Processing an action of type %s", action.action_type)
            action(execution, manager)


@mapper_registry.mapped
class CapabilityRequest(JSONSerializable):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """

    id: int
    state: str
    capability: Capability
    future_products: str
    versions: List[CapabilityVersion]

    __tablename__ = "capability_requests"
    id = sa.Column("capability_request_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))
    sealed = sa.Column("sealed", sa.Boolean, default=False, nullable=False)
    ingested = sa.Column("ingested", sa.Boolean, default=False, server_default="false", nullable=False)
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )
    assigned_da = sa.Column(
        "assigned_da",
        sa.String,
        sa.ForeignKey(QaStaff.user_name),
    )
    assigned_aod = sa.Column(
        "assigned_aod",
        sa.String,
        sa.ForeignKey(QaStaff.user_name),
    )

    # I suspect the trying to set up bidirectional connections for the AoD/DA is going to
    # get confusing, so skipping until desired.
    da = relationship(
        "QaStaff",
        primaryjoin="and_(CapabilityRequest.assigned_da==QaStaff.user_name, QaStaff.group=='DA')",
    )

    aod = relationship(
        "QaStaff",
        primaryjoin="and_(CapabilityRequest.assigned_aod==QaStaff.user_name, QaStaff.group=='AOD')",
    )

    versions = relationship(
        "CapabilityVersion",
        back_populates="request",
        order_by="CapabilityVersion.version_number",
        collection_class=ordering_list("version_number"),
    )
    capability = relationship(Capability, back_populates="requests")

    @property
    def service_url(self):
        if not hasattr(CapabilityRequest, "_service_url"):
            CapabilityRequest._service_url = CapoConfig().settings("edu.nrao.workspaces.UISettings").serviceUrl
        return CapabilityRequest._service_url

    def update_state(self, state: CapabilityRequestState):
        self.state = state.name

    def determine_state(self):
        """
        Determine state of request based on the state of its versions and set it accordingly

        :return:
        """

        CapabilityRequestStateSetter(self).determine_state()

    def update_sealed(self, new_value: bool):
        self.sealed = new_value

    def __str__(self):
        return f"CapabilityRequest object: {self.__dict__}"

    @property
    def created_at_ago(self) -> str:
        return pendulum.instance(self.created_at).diff_for_humans()

    @property
    def updated_at_ago(self) -> str:
        return pendulum.instance(self.updated_at).diff_for_humans()

    @property
    def status_url(self) -> str:
        return self.service_url + "/request-status/" + str(self.id)

    def __repr__(self):
        return (
            f"<CapabilityRequest #{self.id} "
            f"for capability {self.capability_name}, "
            f"created {self.created_at_ago} and "
            f"last updated {self.updated_at_ago})>"
        )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        # Calculate state to ensure it's up to date
        self.determine_state()

        return {
            "type": self.__class__.__name__,
            "id": self.id,
            "capability_name": self.capability_name,
            "state": self.state,
            "assigned_da": self.assigned_da,
            "assigned_aod": self.assigned_aod,
            "sealed": self.sealed,
            "ingested": self.ingested,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "versions": [v.__json__() for v in self.versions],
            "status_url": self.status_url,
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityRequest(**copy)

    @property
    def current_version(self) -> CapabilityVersion:
        return self.versions[-1] if len(self.versions) else None


@mapper_registry.mapped
class CapabilityVersion(JSONSerializable):
    """
    Schema representation of a capability version
    """

    capability_request: CapabilityRequest
    version_number: int
    parameters: List[ParameterIF]
    workflow_metadata: str  # JSON-formatted
    files: List[CapabilityVersionFile]

    __tablename__ = "capability_versions"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
        primary_key=True,
    )
    version_number = sa.Column("version_number", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String, default="Created", server_default="Created")
    parameters = sa.Column("parameters", sa.JSON)
    workflow_metadata = sa.Column("workflow_metadata", sa.JSON)
    sealed = sa.Column("sealed", sa.Boolean, default=False)
    request = relationship(CapabilityRequest, back_populates="versions")
    executions = relationship("CapabilityExecution", back_populates="version")
    files = relationship("CapabilityVersionFile", back_populates="version")
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))
    capability = relationship(Capability)

    @property
    def current_execution(self) -> CapabilityExecution:
        return self.executions[-1] if len(self.executions) else None

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "capability_request_id": self.capability_request_id,
            "version_number": self.version_number,
            "current_execution": self.current_execution.__json__() if self.current_execution else None,
            "state": self.state,
            "parameters": self.parameters,
            "workflow_metadata": self.workflow_metadata,
            "sealed": self.sealed,
            "files": [file.__json__() for file in self.files],
            "capability_name": self.capability.name if self.capability else None,
            "status_url": self.request.status_url,
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        del copy["files"]
        return CapabilityVersion(**copy)


@mapper_registry.mapped
class CapabilityVersionFile(JSONSerializable):
    """
    A Capability Request File is a file that was supplied by the user and attached to the current version
    """

    capability_request_id: int
    version_number: int
    filename: str
    content: bytes

    __tablename__ = "capability_version_files"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        primary_key=True,
    )
    version_number = sa.Column(
        "version_number",
        sa.Integer,
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True, nullable=False)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    version = relationship(CapabilityVersion, back_populates="files")

    __table_args__ = (
        sa.ForeignKeyConstraint(
            [capability_request_id, version_number],
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
        {},
    )

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<CapabilityVersionFile filename={self.filename}>"

    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "capability_request_id": self.capability_request_id,
            "version_number": self.version_number,
            "filename": self.filename,
            "content": self.content.hex(),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return CapabilityVersionFile(filename=json["filename"], content=bytes.fromhex(json["content"]))


class SaPath(sa.types.TypeDecorator):
    impl = sa.types.String

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return pathlib.Path(value)

    def copy(self, **kw):
        return SaPath(self.impl.length)


@mapper_registry.mapped
class CapabilityExecution(JSONSerializable):
    """
    Schema representation of a capability request's execution record
    """

    id: int
    capability_version: CapabilityVersion
    state_name: str
    state: "State"
    version: CapabilityVersion
    current_workflow_request_id: int
    capability: Capability
    capability_request: CapabilityRequest
    delivery_url: str
    delivery_path: pathlib.Path

    __tablename__ = "capability_executions"

    id = sa.Column("execution_id", sa.Integer, primary_key=True)
    state_name = sa.Column("state", sa.String)
    queue_state = sa.Column(
        "queue_state", sa.String, default=QueueState.NotQueued.name, server_default=QueueState.NotQueued.name
    )
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
    )
    version_number = sa.Column("capability_version_number", sa.Integer)
    version = relationship(CapabilityVersion, back_populates="executions")
    current_workflow_request_id = sa.Column("current_workflow_request_id", sa.Integer)
    delivery_url = sa.Column("delivery_url", sa.String)
    delivery_path = sa.Column("delivery_path", SaPath)
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )
    capability = relationship(Capability, viewonly=True)
    state = relationship(State)

    __table_args__ = (
        sa.ForeignKeyConstraint((capability_name, state_name), [State.capability_name, State.state_name]),
        sa.ForeignKeyConstraint(
            (capability_request_id, version_number),
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
    )

    @property
    def capability_request(self):
        return self.version.request

    def __lt__(self, other):
        return self.created_at < other.created_at

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return dict(
            type=self.__class__.__name__,
            id=self.id,
            state_name=self.state_name,
            queue_state=self.queue_state,
            capability_request_id=self.capability_request_id,
            version_number=self.version_number,
            current_workflow_request_id=self.current_workflow_request_id,
            delivery_url=self.delivery_url,
            created_at=self.created_at.isoformat(),
            updated_at=self.updated_at.isoformat(),
        )

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityExecution(**copy)

    def signal(self, manager: ExecutionManagerIF, message: Dict):
        logger.info("Signalling my state %s with this message of type %s", self.state_name, message["type"])
        previous_state = self.state
        # Transition execution state
        self.state = self.state.get_next_state(message)
        # Perform action
        previous_state.perform_action(message, self, manager)


class CapabilityRequestStateSetter:
    """Sets request state based on state(s) of its version(s)"""

    def __init__(self, request: CapabilityRequest):
        self.request = request
        self.version_states = [version.state for version in self.request.versions if version.state]

    def determine_state(self):
        """
        Determine state of request based on state(s) of its version(s).

        RULES:
        - If -any- version is Complete, and -all- other versions are either Failed or Complete, the request is Complete
        - If all versions are Failed, the request is Failed
        - If all versions are Complete, the request is Complete
        - If all versions are Created, the request is Created
        - Otherwise, it is Submitted

        Because capabilities depend on the stringified CapabilityRequestState,
        here we set the request state to str rather than to state object.

        :return:
        """

        if not self.version_states:
            return

        if self.any_version_complete():
            if self.all_versions_same() or self.any_complete_all_others_failed():
                self.request.state = CapabilityRequestState.Complete.name
        elif self.all_versions_in_state(CapabilityVersionState.Failed.name) or self.all_versions_in_state(
            CapabilityVersionState.Error.name
        ):
            self.request.state = CapabilityRequestState.Failed.name
        elif self.any_version_cancelled():
            self.request.state = CapabilityRequestState.Cancelled.name
        elif self.all_versions_in_state(CapabilityVersionState.Created.name):
            self.request.state = CapabilityRequestState.Created.name
        elif self.all_versions_in_state(CapabilityVersionState.Running.name):
            self.request.state = CapabilityRequestState.Submitted.name
        elif self.all_versions_in_state(CapabilityVersionState.Cancelled.name):
            self.request.state = CapabilityRequestState.Cancelled.name
        else:
            self.evaluate_mixed_versions()

    def evaluate_mixed_versions(self):
        """
        Determine request state if no version is either Failed or Complete.

        :return:
        """
        if CapabilityVersionState.Running.name in self.version_states:
            self.request.state = CapabilityRequestState.Submitted.name
        elif CapabilityVersionState.Created.name in self.version_states:
            self.request.state = CapabilityRequestState.Created.name
        elif CapabilityVersionState.Cancelled.name in self.version_states:
            self.request.state = CapabilityRequestState.Cancelled.name

    def all_versions_in_state(self, state: str) -> bool:
        """
        Given a request with multiple versions, is every version in this state?

        :param state:
        :return:
        """
        if self.all_versions_same():
            return self.request.versions[0].state == state

    def all_versions_same(self) -> bool:
        """
        We've got a request with more than one version.
        Are all the versions' states identical?

        :return:
        """
        return len(set(self.version_states)) == 1

    def any_version_complete(self) -> bool:
        """
        Of all the versions in this request, is at least one of them complete?

        :return:
        """

        return any(state == CapabilityVersionState.Complete.name for state in self.version_states)

    def any_version_cancelled(self) -> bool:
        """
        Of all the versions in this request, is at least one of them Cancelled?

        :return:
        """

        return any(state == CapabilityVersionState.Cancelled.name for state in self.version_states)

    def any_complete_all_others_failed(self) -> bool:
        """
        Given a request with at least one Complete version, are all the others Failed?

        :return:
        """
        if self.any_version_complete():
            for state in self.version_states:
                if state not in [
                    CapabilityVersionState.Cancelled.name,
                    CapabilityVersionState.Failed.name,
                    CapabilityVersionState.Error.name,
                    CapabilityVersionState.Complete.name,
                ]:
                    return False
            return True
        return False
