#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" All The Things about Capabilities """

# pylint: disable=C0103, C0301, C0303, E0401, R0205, R0903, R0904, R0913, W1203

import json
import logging
import subprocess
import sys
from datetime import datetime
from typing import Dict, List, Optional

import transaction
from sqlalchemy import desc, distinct, text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
    CapabilityVersionFile,
    QaStaff,
)
from workspaces.capability.services.interfaces import QueueReport
from workspaces.capability.services.restriction_engine import (
    RestrictionEngine,
    Restrictions,
)
from workspaces.products.schema_interfaces import FutureProductIF

Base = declarative_base()

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class CapabilityName(object):
    """Encapsulates the name of a capability"""


class CapabilityInfo:
    """
    Interface for data access object that can look up and record information about capabilities
    and capability requests. Accesses and modifies the database.
    """

    def __init__(self, session: Session):
        self.session = session
        self.restriction_engine = RestrictionEngine(self)

    def create_capability(
        self,
        name: CapabilityName,
        max_jobs: int,
        has_image_products: bool,
        requires_qa: bool,
        single_version_only: bool,
    ) -> Capability:
        """
        Create new capability and save it in the database

        :param name: Name of new capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :param has_image_products:  Does this capability produce images?
        :param requires_qa: Does this capability require QA?
        :param single_version_only: Does this capability allow for only one version per request?
        :return: Created capability
        """
        capability = Capability(
            name=name,
            max_jobs=max_jobs,
            has_image_products=has_image_products,
            requires_qa=requires_qa,
            single_version_only=single_version_only,
        )
        self.save_entity(capability)
        return capability

    def edit_capability(
        self,
        name: CapabilityName,
        max_jobs: int = None,
        enabled: bool = None,
        has_image_products: bool = None,
        requires_qa: bool = None,
        single_version_only: bool = None,
    ) -> bool:
        """
        Edit existing capability definition

        :param name: Name of capability to edit
        :param max_jobs: New number of max jobs or None if wanting to leave unchanged
        :param enabled: New enabled state or None if wanting to leave unchanged
        :param has_image_products: New has_image_products value or None if wanting to leave unchanged
        :param requires_qa: New requires_qa state or None if wanting to leave unchanged
        :param single_version_only: New single_version_only value or None if wanting to leave unchanged
        :return: True if the capability was successfully edited, else False
        """
        changes = {}

        if max_jobs is not None:
            changes[Capability.max_jobs] = max_jobs
        if enabled is not None:
            changes[Capability.enabled] = enabled
        if has_image_products is not None:
            changes[Capability.has_image_products] = has_image_products
        if requires_qa is not None:
            changes[Capability.requires_qa] = requires_qa

        if changes:
            rows_changed = self.session.query(Capability).filter_by(name=name).update(changes)
        else:
            logger.info("No changes were made, because none were given.")
            return True

        if rows_changed == 1:
            self.session.flush()
            return True
        else:
            return False

    def pause_capability(self, capability_name: str):
        """
        Pause capability processing, preventing future requests from being executed

        :param capability_name: Name of capability to pause
        """
        capability = self.lookup_capability(capability_name)
        capability.paused = True
        self.save_entity(capability)

    def unpause_capability(self, capability_name: str):
        """
        Unpause capability processing, allowing future requests to be executed normally

        :param capability_name: Name of capability to unpause
        """
        capability = self.lookup_capability(capability_name)
        capability.paused = False
        self.save_entity(capability)
        transaction.commit()

    @staticmethod
    def get_metadata_from_wrester(parameters: List) -> List:
        """
        Run AAT Wrest to get Observation metadata. Only needs to occur on first version creation

        :param parameters:
        :return: a dictionary containing metadata needed for template rendering
        """
        metadata = {}
        wrest_type = argument = None
        if parameters is not None:
            if "product_locator" in parameters:
                if parameters["product_locator"] is not None:
                    wrest_type = "--observation-via-spl"
                    argument = parameters["product_locator"]
            elif "sdmId" in parameters:
                if parameters["sdmId"] is not None:
                    wrest_type = "--observation"
                    argument = parameters["sdmId"]
                    metadata["sdm_id"] = parameters["sdmId"]
        else:
            logger.info(f"No parameters provided")

        if argument and wrest_type:
            with subprocess.Popen(
                ["aat_wrest", wrest_type, argument],
                stdout=subprocess.PIPE,
            ) as proc:
                process_json = proc.communicate()[0].decode()
                logger.info(process_json)
                if "ERROR" not in process_json:
                    metadata.update(json.loads(process_json))
                else:
                    logger.error(process_json)
        return metadata

    @staticmethod
    def run_wrester(capability_request: CapabilityRequest) -> Dict:
        """
        Run AAT Wrest for template generation. Only needs to occur on first version creation

        :param capability_request: the request and initial version to render templates for
        :return: a dictionary containing metadata needed for template rendering
        """
        name = capability_request.capability_name
        version = capability_request.current_version
        if "calibration" in name:
            wrest_type = "-sc"
            argument = version.parameters["product_locator"]
            argument2 = []
        elif "restore" in name:
            wrest_type = "-r"
            argument = version.parameters["product_locator"]
            argument2 = [version.parameters["cal_locator"]]
        elif "imaging" in name:
            wrest_type = "-si"
            argument = version.parameters["sdmId"]
            argument2 = []
        else:
            logger.info(f"No wrester found for capability {name}. Assuming no templates needed.")
            return {}

        logger.info(f"capability {name} has wrest option: {wrest_type}")
        with subprocess.Popen(
            ["aat_wrest", wrest_type, argument] + argument2,
            stdout=subprocess.PIPE,
        ) as metadata:
            process_json = metadata.communicate()[0].decode()
            logger.info(process_json)
            if "ERROR" not in process_json:
                return json.loads(process_json)
            else:
                logger.error(process_json)
                return {}

    def render_capability_templates(self, capability_request: CapabilityRequest):
        """
        Render and save all editable template files for a *new* Capability Request.
        Subsequent versions will pick up the first rendered templates from the previous version

        :param capability_request: The request with initial version to render templates for
        """
        version = capability_request.current_version

        wrested_args = self.run_wrester(capability_request)

        if wrested_args:
            arguments = {
                **wrested_args,
                **version.parameters,
                "relative_path": "Waiting for HTCondor",
                "root_directory": "Waiting for HTCondor",
            }

            templated_files = capability_request.capability.render_templates(arguments)
            for file in templated_files:
                self.save_version_file(version=version, filename=file.filename, content=file.content)

    def create_capability_request(
        self,
        capability_name: str,
        parameters: List[ParameterIF] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequest:
        """
        Create new capability request and save it in the database

        :param capability_name: name of the requested capability
        :param parameters: List
        :param future_products:
        :return: Created CapabilityRequest
        """
        capability = self.lookup_capability(capability_name)
        if not any(item in capability.name for item in ["null", "download"]):
            metadata = self.get_metadata_from_wrester(parameters)
            if parameters is not None:
                if "metadata" not in parameters:
                    if isinstance(parameters, List):
                        parameters.append(metadata)
                    else:
                        parameters["metadata"] = metadata
                else:
                    for key in metadata:
                        parameters["metadata"][key] = metadata[key]

                if "product_locator" not in parameters:
                    parameters["product_locator"] = metadata["spl"]

        request = CapabilityRequest(
            state=CapabilityRequestState.Created.name,
            capability=capability,
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=parameters, capability_name=capability_name)],
        )
        self.save_entity(request)

        self.render_capability_templates(request)
        return request

    def create_capability_version(
        self,
        capability_request_id: int,
        parameters: List[ParameterIF] = None,
    ) -> CapabilityVersion:
        """
        Create new capability version and save it to the database

        :param capability_request_id: id of the request to attach new version to
        :param parameters:
        :return: new CapabilityVersion
        """

        request = self.lookup_capability_request(capability_request_id)

        version = self.restriction_engine.trigger(request, Restrictions.CREATE_VERSION, trigger_args=parameters)
        return version

    def create_execution(self, version: CapabilityVersion) -> CapabilityExecution:
        """
        Create new execution record for a request version and save it in the database

        :param version: Capability request version that needs an execution
        :return: Created CapabilityExecution
        """
        execution = self.restriction_engine.trigger(version, Restrictions.CREATE_EXECUTION)
        return execution

    def lookup_capability(self, capability_name: str) -> Capability:
        return self.session.query(Capability).filter_by(name=capability_name).first()

    def lookup_capability_request(self, request_id) -> CapabilityRequest:
        """
        Finds the capability request with the provided request ID
        :param request_id: capability request ID of interest

        :return:
        """
        return self.session.query(CapabilityRequest).filter_by(id=request_id).first()

    def get_enabled_capabilities(self) -> List[Capability]:
        """
        Gets the names of Capabilities that are enabled

        :return: List of names of enabled Capabilities
        """
        list_of_enabled_capabilities = self.session.query(Capability).filter(Capability.enabled).all()
        json_list = []
        for request in list_of_enabled_capabilities:
            json_list.append(request.name)
        return {"enabled_capabilities": json_list}

    def lookup_active_capability_requests(
        self, capability_name: str, filters: dict
    ) -> Dict[str, List[CapabilityRequest]]:
        """
        Find all the active capability requests with this name.

        :param capability_name:
        :return:
        """
        default_query = (
            self.session.query(CapabilityRequest)
            .filter_by(capability_name=capability_name)
            .filter(CapabilityRequest.state != "Complete")
            .filter(CapabilityRequest.sealed == False)
            .order_by(CapabilityRequest.id.desc())
        )
        if filters.__contains__("state"):
            default_query = default_query.filter(CapabilityRequest.state.in_(filters.get("state")))
        if filters.__contains__("assigned_da"):
            default_query = default_query.filter(CapabilityRequest.assigned_da.in_(filters.get("assigned_da")))
        if filters.__contains__("assigned_aod"):
            default_query = default_query.filter(CapabilityRequest.assigned_aod.in_(filters.get("assigned_aod")))
        if filters.__contains__("date_observed"):
            default_query = default_query.filter(CapabilityRequest.state.in_(filters.get("state")))

        active_list = default_query.all()

        json_list = []
        for request in active_list:
            json_list.append(request.__json__())

        return {"active_requests": json_list}

    def lookup_created_capability_requests(self, capability_name: str) -> Dict[str, List[CapabilityRequest]]:
        """
        Find all the inactive (created but not submitted) capability requests with this name.

        :param capability_name: Name of capability
        :return: List of created requests as a dict
        """
        active_list = (
            self.session.query(CapabilityRequest)
            .filter_by(capability_name=capability_name)
            .filter(CapabilityRequest.state == "Created")
            .all()
        )

        json_list = []
        for request in active_list:
            json_list.append(request.__json__())

        return {"created_requests": json_list}

    def lookup_version(self, request_id: int, version_number: int) -> Optional[CapabilityVersion]:
        """
        Return version with the given version number for the given capability request

        :param request_id: ID of request
        :param version_number: Version number
        :return: Version with that number or None if the query doesn't find anything
        """
        return (
            self.session.query(CapabilityVersion)
            .filter_by(capability_request_id=request_id, version_number=version_number)
            .one()
        )

    def lookup_version_file(self, request_id: int, version_number: int, filename: str) -> Optional[CapabilityVersion]:
        """
        Returns file with the given request ID, version number, and filename for the given capability request

        :param request_id: ID of request
        :param version_number: Version number
        :param filename: name of file
        :return: Version with that number or None if the query doesn't find anything
        """
        return (
            self.session.query(CapabilityVersionFile)
            .filter_by(capability_request_id=request_id, version_number=version_number, filename=filename)
            .one()
        )

    def lookup_current_version(self, request_id) -> CapabilityVersion:
        """
        What's the current capability version for this request ID?

        :param request_id:
        :return:
        """
        return (
            self.session.query(CapabilityVersion)
            .filter_by(capability_request_id=request_id)
            .order_by(desc("version_number"))
            .first()
        )

    def lookup_execution(self, execution_id: int) -> CapabilityExecution:
        """
        Finds the capability execution with the provided request ID

        :param execution_id: capability execution ID of interest
        :return: the CapabilityExecution we want
        """
        return self.session.query(CapabilityExecution).filter_by(id=execution_id).first()

    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()

    def lookup_execution_by_workflow_request_id(self, workflow_request_id: int) -> CapabilityExecution:
        """
        Find execution record given a capability request

        :param workflow_request_id:
        :return: Capability execution record
        """
        return (
            self.session.query(CapabilityExecution).filter_by(current_workflow_request_id=workflow_request_id).first()
        )

    def lookup_execution_by_request_version(self, request_id: int, version_number: int) -> CapabilityExecution:
        """
        Find execution record given a capability request version

        :param request_id
        :param version_number
        :return: Capability execution record
        """
        return (
            self.session.query(CapabilityExecution)
            .filter_by(capability_request_id=request_id, version_number=version_number)
            .one_or_none()
        )

    def requests_for_capability(self, capability_name: str) -> List[CapabilityRequest]:
        """
        Gets all capability requests for a given capability

        :param capability_name: Name of capability
        :return: List of associated requests
        """
        return self.session.query(CapabilityRequest).filter_by(capability_name=capability_name).all()

    def save_execution(self, execution: CapabilityExecution):
        """
        Commit this execution.

        :param execution:
        """
        logger.info(
            "Saving execution %s with state %s",
            execution.id,
            execution.state,
        )
        self.session.add(execution)
        self.session.flush()

    def save_version(self, version: CapabilityVersion):
        """
        Commit this version

        :param version: the version to save
        """
        logger.info(f"Saving version {version.version_number} of request {version.capability_request_id}")
        self.session.add(version)
        self.session.flush()

    def save_request(self, request: CapabilityRequest):
        """
        Commit this request.

        :param request: Capability request to commit
        """
        logger.info("Saving request %s", request.id)
        self.session.add(request)
        self.session.flush()

    def save_version_file(self, version: CapabilityVersion, filename: str, content: bytes) -> CapabilityVersionFile:
        """
        Save a given file for the specified request version to the database

        :param version: request version to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return:
        """
        vf = CapabilityVersionFile(
            capability_request_id=version.capability_request_id,
            version_number=version.version_number,
            filename=filename,
            content=content,
        )
        self.session.add(vf)
        self.session.flush()
        return vf

    def save_edited_version_file(self, capabilityVersionFile: CapabilityVersionFile) -> CapabilityVersionFile:
        """
        Save a given edited file to the database

        :param capabilityVersionFile: edited version file to save to the database
        :return: the edited capability version file
        """

        self.session.add(capabilityVersionFile)
        self.session.flush()
        return capabilityVersionFile

    def copy_previous_version_files(
        self, version: CapabilityVersion, previous_version_files: List[CapabilityVersionFile]
    ):
        """
        Saves previous version files.

        :param version: version of interest
        :param previous_version_files: version files to save

        :return:
        """
        logger.info(f"FILES: {previous_version_files}")
        for file in previous_version_files:
            self.save_version_file(version=version, filename=file.filename, content=file.content)

    def delete_request_versions(self, request_id: int) -> int:
        """
        Deleting all versions of a capability request. This is executed as preparation for deleting the request itself

        :param request_id: ID of request that's versions will be deleted
        :return: Number of rows deleted
        """
        rows_changed = self.session.query(CapabilityVersion).filter_by(capability_request_id=request_id).delete()
        self.session.flush()
        return rows_changed

    def delete_capability_request(self, request_id: int) -> int:
        """
        Delete a specified capability request

        :param request_id: ID of capability request to delete
        :return: Number of rows deleted
        """
        if self.can_delete_capability_request(request_id):
            self.delete_request_versions(request_id)
            rows_changed = self.session.query(CapabilityRequest).filter_by(id=request_id).delete()
            self.session.flush()
            return rows_changed

        return -1

    def can_delete_capability_request(self, request_id: int) -> bool:
        """
        Method that checks if a capability request has executions associated with it. If it does, it cannot be deleted.

        :param request_id: ID of capability request to be checked
        :return: True if there are no executions associated with request; else False
        """
        if self.session.query(CapabilityExecution).filter_by(capability_request_id=request_id).all():
            return False

        return True

    def get_available_qa_staff_in_group(self, group: str) -> List[QaStaff]:
        """
        Retrieve unique list of QA reviewer groups

        :return: the groups
        """
        qa_staff = (
            self.session.query(QaStaff)
            .order_by("user_name")
            .filter(QaStaff.group == group)
            .filter(QaStaff.available == True)
            .all()
        )
        json_staff = []
        for staff in qa_staff:
            json_staff.append(staff.__json__())
        return json_staff

    @staticmethod
    def get_staff_member(json: dict) -> QaStaff:
        return QaStaff.from_json(json)

    # ---------------------------------------
    # Methods related to the QueueReporterIF
    # ---------------------------------------

    def report(self) -> List[QueueReport]:
        result = []
        for row in self.session.get_bind().execute(
            text(
                """
                with
                    queue_states as (
                        select cr.capability_name, ce.queue_state, count(*) as in_state
                        from capability_executions ce
                             join capability_versions cv on ce.capability_request_id = cv.capability_request_id and ce.capability_version_number = cv.version_number
                             join capability_requests cr on cv.capability_request_id = cr.capability_request_id
                        where ce.queue_state = 'Queued'
                        group by cr.capability_name, ce.queue_state
                        union all
                        select cr.capability_name, ce.queue_state, count(*) as in_state
                        from capability_executions ce
                             join capability_versions cv on ce.capability_request_id = cv.capability_request_id and ce.capability_version_number = cv.version_number
                             join capability_requests cr on cv.capability_request_id = cr.capability_request_id
                        where ce.queue_state = 'Running'
                        group by cr.capability_name, ce.queue_state)
                SELECT
                    capability_name,
                    COALESCE((SELECT in_state FROM queue_states WHERE c.capability_name = queue_states.capability_name AND queue_state = 'Queued'), 0) AS waiting,
                    COALESCE((SELECT in_state FROM queue_states WHERE c.capability_name = queue_states.capability_name AND queue_state = 'Running'), 0) AS running,
                    max_jobs
                FROM capabilities c WHERE NOT c.paused"""
            )
        ):
            result.append(QueueReport(row[0], row[1], row[2], row[3]))
        return result

    def waiting_executions(self, queue: str, count: int) -> List[CapabilityExecution]:
        # We obtain here the first execution in this queue (capability) with state Submitted,
        # meaning it has not yet been sent to be executed.
        #
        # We assume here that the priority of an execution is just first-in, first-out.
        # Ordering by the ID like this is acting as a kind of proxy for the temporal
        # fact we need of when they were created.
        # TODO: This will need to be revisited when we implement priority queues.
        result = self.session.get_bind().execute(
            text(
                """
            select distinct ce.execution_id
            from capability_executions ce
            join capability_versions cv on ce.capability_request_id = cv.capability_request_id and
                                           ce.capability_version_number = cv.version_number
            join capability_requests cr on cv.capability_request_id = cr.capability_request_id
            where ce.queue_state = 'Queued' and
                  cr.capability_name = :queue
            order by ce.execution_id
            limit :count"""
            ),
            queue=queue,
            count=count,
        )
        return [self.lookup_execution(r[0]) for r in result]


class RestrictedInfo(CapabilityInfo):
    """
    Class for info methods that are restricted.
    """

    def create_version(
        self,
        request: CapabilityRequest,
        parameters: List[ParameterIF],
    ) -> CapabilityVersion:
        # Reset request state accordingly
        request.determine_state()
        self.save_entity(request)
        logger.info(f"Parent Request: {request.__json__()}")

        current_version_files = request.current_version.files

        version = CapabilityVersion(
            capability_request_id=request.id,
            version_number=len(request.versions) + 1,
            parameters=parameters,
            request=request,
            capability=request.capability,
        )
        self.save_entity(version)
        logger.info(f"New Version: {version.__json__()}")

        self.copy_previous_version_files(version=version, previous_version_files=current_version_files)
        return version

    def create_execution(self, version: CapabilityVersion):
        # look up the most recent version
        execution = CapabilityExecution(
            state_name=version.request.capability.start_state,
            queue_state=QueueState.NotQueued.name,
            version=version,
            # ensure that we have a copy of the step sequence as it was when the execution started
            created_at=datetime.now(),
            updated_at=datetime.now(),
            capability_name=version.capability_name,
        )
        self.save_entity(execution)
        return execution

    def create_workflow(
        self,
        execution: CapabilityExecution,
        args: dict,
    ):
        workflow_service = args[0]
        workflow_name = args[1]
        args_list = args[2]
        files = args[3]

        logger.info(
            "creating workflow request with name %s, args %s (and arg type %s)",
            workflow_name,
            args_list,
            type(args_list),
        )
        workflow_request = workflow_service.create_workflow_request(
            workflow=workflow_name,
            argument=args_list,
        )

        logger.info(
            "My workflow_request is %s, with type %s and workflow request ID %s",
            workflow_request,
            type(workflow_request),
            workflow_request.workflow_request_id,
        )

        for file in files:
            workflow_service.attach_file_to_request(workflow_request, file.filename, file.content)
            logger.info(f"Added file {file.filename} to workflow request #{workflow_request.workflow_request_id}")

        execution.current_workflow_request_id = workflow_request.workflow_request_id
        self.save_execution(execution)

        return workflow_request

    @staticmethod
    def edit_parameters(version: CapabilityVersion):
        # TODO: Not yet implemented in system
        pass

    @staticmethod
    def edit_files(version: CapabilityVersion):
        # TODO: Not yet implemented in system
        pass
